#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <fcntl.h>
#include <linux/if.h>
#include <linux/if_ppp.h>
#include <linux/ppp_defs.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <syslog.h>
#include <utime.h>
#include "common.h"

#define		MAX_BOOT_WARN_TIMES		16
#define		MAX_PPP_UNITS			256
#define		MSG_BUSY			0
#define		MSG_IDLE			1
#define		MSG_NO_TIME			2

void		daemonInit();
void		processConfig();
int		update(int sig);
char		*userTTY(char *devName, int uid);
void		userMessage(char *login, char *devName, char *format, ...);
int		devNumFromName(char *devName);
int		PPPidleLeft(UserRec *ur, char *ttyPath);
void		PPPupdate(UserRec *ur, int unit);
void		PPPfindUnits();
int		PPPunit(uid_t uid);
void		logQuit(char *format, ...);

const char	*msgTimeWarning = "You have %d minute\\(s\\) remaining.";
const char	*msgIdleWarning = "You will be logged out after 1 more minute of idle time.";
const char	*msgPath[3] = { LIB"/user_updated.busy", LIB"/user_updated.idle", LIB"/user_updated.no_time" };

time_t		configTime;
int		optSendMail;
int		optGuestTime, optGuestPriority;
int		optSmartBoot, optSmartBootOne;
int		allBusy;
int		multiBase;
int		nLines;
int		optIdleBoot;
int		nBootWarnTimes;
int		bootWarnTime[MAX_BOOT_WARN_TIMES];
int		nLogins;
LoginRec	loginRec[MAX_USERS];
int		nProcs;
ProcRec		procRec[MAX_PROCS];
uid_t		guestUID;
pid_t		PPPuser[MAX_PPP_UNITS];

void main() {
  struct stat _stat;
  time_t t;
  struct tm *tmP;
  daemonInit();
  openlog("user_updated", 0, LOG_DAEMON);
  guestUID = uidFromLogin("guest");
  for (;;) {
    if (stat(configFilePath, &_stat) < 0)
      logQuit("couldn't stat %s", configFilePath);
    if (_stat.st_mtime > configTime) {
      processConfig();
      if (configTime)
        syslog(LOG_INFO, "re-read %s", configFilePath);
      configTime = _stat.st_mtime;
    }
    userList(&nLogins, loginRec);
    procList(&nProcs, procRec);
    PPPfindUnits();
    allBusy = allLinesBusy(nProcs, procRec, multiBase, nLines);
    if (update(SIGTERM)) {
      sleep(5);
      userList(&nLogins, loginRec);
      procList(&nProcs, procRec);
      update(SIGKILL);
    }
    t = time(NULL);
    tmP = localtime(&t);
    sleep(60 - tmP->tm_sec);
  }
}

void daemonInit() {
  int i;
  if (fork()) exit(0);
  for (i = 0; i <= 2; i++)
    close(i);
  setsid();
  chdir("/");
  umask(0);
}

void processConfig() {
  int li;
  char line[256], token[256], devPath[UT_LINESIZE];
  FILE *f;
  struct stat _stat;
  if ((f = fopen(configFilePath, "rt")) == NULL)
    logQuit("couldn't open config file");
  optSendMail = 0;
  optGuestTime = 15;
  optGuestPriority = 4;
  nBootWarnTimes = 0;
  optIdleBoot = 0;
  optSmartBoot = 0;
  optSmartBootOne = 0;
  while (fgets(line, 256, f) != NULL) {
    li = 0;
    if (getToken(token, line, &li)) continue;
    if (!strcasecmp(token, "Devices")) {
      if (getToken(token, line, &li)) continue;
      sprintf(devPath, "/dev/%s", token);
      if (stat(devPath, &_stat) < 0)
	perrQuit("stat");
      multiBase = _stat.st_rdev;
      if (getToken(token, line, &li)) continue;
      nLines = atoi(token);
    } else if (!strcasecmp(token, "SendMail")) {
      optSendMail = 1;
    } else if (!strcasecmp(token, "GuestTime")) {
      if (getToken(token, line, &li)) continue;
      optGuestTime = atoi(token);
    } else if (!strcasecmp(token, "GuestPriority")) {
      if (getToken(token, line, &li)) continue;
      optGuestPriority = atoi(token);
    } else if (!strcasecmp(token, "WarnBoot")) {
      while (!getToken(token, line, &li) && nBootWarnTimes < MAX_BOOT_WARN_TIMES)
	bootWarnTime[nBootWarnTimes++] = atoi(token);
    } else if (!strcasecmp(token, "IdleBoot")) {
      if (getToken(token, line, &li)) continue;
      optIdleBoot = atoi(token);
    } else if (!strcasecmp(token, "SmartBoot")) {
      optSmartBoot = 1;
    } else if (!strcasecmp(token, "SmartBootOne")) {
      optSmartBootOne = 1;
    }
  }
  fclose(f);
}

int update(int sig) {
  char login[UT_NAMESIZE + 1], devName[UT_LINESIZE], str[256];
  int minPriority = 8, minSleft = 1, zeroMinSleft = 0x7fffffff, zeroBoot = 0;
  int i, j, k, unit, idleLeft, sLeft, doKill, guestOnline;
  struct stat _stat;
  UserRec ur;
  int nVictims = 0, nKill = 0;
  uid_t victim[256];
  byte victimMsg[256];
  pid_t	killPID[MAX_PROCS];
  for (guestOnline = 0, i = 0; i < nProcs; i++)
    if (procRec[i].uid == guestUID) guestOnline = 1;
  userFileOpen();
  while (!userFileRead(&ur)) {
    for (i = 0; i < nProcs && ur.uid != procRec[i].uid; i++);
    if (i == nProcs) continue;
    loginFromUID(login, ur.uid);
    if (!userTTY(devName, procRec[i].uid)) continue;
    unit = PPPunit(ur.uid);
    if (optIdleBoot) {
      if (unit >= 0) idleLeft = PPPidleLeft(&ur, devName);
      else {
	if (stat(devName, &_stat)) continue;
	idleLeft = optIdleBoot * 60 - (time(NULL) - _stat.st_atime);
      }
      if (idleLeft < 0) idleLeft = 0;
      else idleLeft = idleLeft / 60 + 1;
    } else idleLeft = 0x7fffffff;
    if (sig == SIGTERM) {
      ur.tLeft--;
      ur.sLeft--;
      time(&ur.last);
      if (unit >= 0) PPPupdate(&ur, unit);
      else {
	if (idleLeft == 1 && ur.sLeft > 1) 
	  userMessage(login, devName + 5, (char*)msgIdleWarning);
	else for (j = 0; j < nBootWarnTimes; j++)
	  if (ur.sLeft == bootWarnTime[j]) {
	    userMessage(login, devName + 5, (char*)msgTimeWarning, ur.sLeft);
	    break;
	  }
      }
      userFileEdit(&ur);
    }
    if (optSmartBoot) {
      if (allBusy) {
	if (idleLeft <= 0 || ur.sLeft <= 0) {
	  if (optSmartBootOne) {
	    if (PRIORITY(ur.flags) < minPriority || (PRIORITY(ur.flags) == minPriority && ur.sLeft < minSleft)) {
	      minPriority = PRIORITY(ur.flags);
	      minSleft = ur.sLeft;
	      zeroBoot = 0;
	      nKill = 0;
	      nVictims = 0;
	      if (optSendMail && unit >= 0) {
		victim[nVictims] = ur.uid;
		victimMsg[nVictims++] = ur.sLeft <= 0 ? MSG_NO_TIME : MSG_IDLE;
	      }
	      doKill = 1;
	    } else doKill = 0;
	  } else {
	    if (zeroBoot) {
	      zeroBoot = 0;
	      nKill = 0;
	      nVictims = 0;
	    }
	    if (optSendMail && unit >= 0) {
	      victim[nVictims] = ur.uid;
	      victimMsg[nVictims++] = ur.sLeft <= 0 ? MSG_NO_TIME : MSG_IDLE;
	    }
	    doKill = 1;
	  }
	} else if (!PRIORITY(ur.flags) && allBusy && (!nKill || zeroBoot) && ur.sLeft < zeroMinSleft) {
	  zeroMinSleft = ur.sLeft;
	  zeroBoot = 1;
	  nKill = 0;
	  nVictims = 0;
	  if (optSendMail && unit >= 0) {
	    victim[nVictims] = ur.uid;
	    victimMsg[nVictims++] = MSG_BUSY;
	  }
	  doKill = 1;
	} else doKill = 0;
      } else doKill = 0;
    } else {
      if (idleLeft <= 0 || ur.sLeft <= 0) {
	if (zeroBoot) {
	  zeroBoot = 0;
	  nKill = 0;
	  nVictims = 0;
	}
	if (optSendMail && unit >= 0) {
	  victim[nVictims] = ur.uid;
	  victimMsg[nVictims++] = ur.sLeft <= 0 ? MSG_NO_TIME : MSG_IDLE;
	}
	doKill = 1;
      } else if (!PRIORITY(ur.flags) && allBusy && (!nKill || zeroBoot) && ur.sLeft < zeroMinSleft) {
	zeroMinSleft = ur.sLeft;
	zeroBoot = 1;
	nKill = 0;
	nVictims = 0;
	if (optSendMail && unit >= 0) {
	  victim[nVictims] = ur.uid;
	  victimMsg[nVictims++] = MSG_BUSY;
	}
	doKill = 1;
      } else doKill = 0;
    }
    if (doKill)
      for (; i < nProcs; i++)
	if (procRec[i].uid == ur.uid) killPID[nKill++] = procRec[i].pid;
  }
  userFileClose();
  for (i = 0; i < nLogins; i++) {
    if (loginRec[i].uid != guestUID) continue;
    if (optIdleBoot) {
      if (stat(loginRec[i].tty, &_stat)) continue;
      idleLeft = optIdleBoot * 60 - (time(NULL) - _stat.st_atime);
      if (idleLeft < 0) idleLeft = 0;
      else idleLeft = idleLeft / 60 + 1;
    } else idleLeft = 0x7fffffff;
    sLeft = (optGuestTime * 60 - (time(NULL) + (sig == SIGTERM ? 5 : 0) - loginRec[i].time)) / 60 + 1;
    if (sig == SIGTERM) {
      if (idleLeft == 1 && sLeft > 1)
	userMessage("guest", loginRec[i].tty, (char*)msgIdleWarning);	
      else for (j = 0; j < nBootWarnTimes; j++)
        if (sLeft == bootWarnTime[j]) {
 	  userMessage("guest", loginRec[i].tty, (char*)msgTimeWarning, sLeft);
	  break;
        }
    }
    if (optSmartBoot) {
      if (allBusy) {
	if (idleLeft <= 0 || sLeft <= 0) {
	  if (optSmartBootOne) {
	    if (optGuestPriority < minPriority || (optGuestPriority == minPriority && sLeft < minSleft)) {
	      minSleft = sLeft;
	      zeroBoot = 0;
	      nKill = 0;
	      nVictims = 0;
	      doKill = 1;
	    } else doKill = 0;
	  } else {
	    if (zeroBoot) {
	      zeroBoot = 0;
	      nKill = 0;
	      nVictims = 0;
	    }
	    doKill = 1;
	  }
	} else if (!optGuestPriority && allBusy && (!nKill || zeroBoot) && sLeft < zeroMinSleft) {
	  zeroMinSleft = sLeft;
	  zeroBoot = 1;
	  nKill = 0;
	  nVictims = 0;
	  doKill = 1;
	} else doKill = 0;
      } else doKill = 0;
    } else {
      if (idleLeft <= 0 || sLeft <= 0) {
	if (zeroBoot) {
	  zeroBoot = 0;
	  if (optSendMail) nVictims = 0;
	  nKill = 0;
	}
	doKill = 1;
      } else if (!optGuestPriority && allBusy && (!nKill || zeroBoot) && sLeft < zeroMinSleft) {
	zeroMinSleft = sLeft;
	zeroBoot = 1;
	nKill = 0;
	nVictims = 0;
	doKill = 1;
      } else doKill = 0;
    }
    if (doKill) {
      k = devNumFromName(loginRec[i].tty);
      for (j = 0; j < nProcs; j++)
	if (procRec[j].tty == k) killPID[nKill++] = procRec[j].pid;
    }
  }
  for (i = 0; i < nKill; i++) kill(killPID[i], sig);
  if (optSendMail && sig == SIGTERM)
    for (i = 0; i < nVictims; i++) {
      loginFromUID(login, victim[i]);
      sprintf(str, "mail -s \"Connection Terminated\" %s < %s", login, msgPath[victimMsg[i]]);
      system(str);
    }
  return nKill;
}

char *userTTY(char *devName, int uid) {
  int i;
  for (i = 0; i < nLogins; i++)
    if (loginRec[i].uid == uid) {
      sprintf(devName, "/dev/%s", loginRec[i].tty);
      return devName;
    }
  devName[0] = '\0';
  return NULL;
}

void userMessage(char *login, char *devName, char *format, ...) {
  char tty[UT_LINESIZE], buf[256], line[256];
  struct stat _stat;
  struct utimbuf _utime;
  va_list ap;
  va_start(ap, format);
  if (optIdleBoot) {
    sprintf(tty, "/dev/%s", devName);
    stat(tty, &_stat);
  }
  vsprintf(buf, format, ap);
  sprintf(line, "echo %s | write %s %s", buf, login, devName);
  system(line);
  if (optIdleBoot) {
    _utime.actime = _utime.modtime = _stat.st_atime;
    utime(tty, &_utime);
  }
  va_end(ap);
}

int devNumFromName(char *devName) {
  struct stat _stat;
  char devPath[UT_LINESIZE];
  sprintf(devPath, "/dev/%s", devName);
  if (stat(devPath, &_stat) < 0)
    perrQuit("stat");
  return _stat.st_rdev;
}

int PPPidleLeft(UserRec *ur, char *ttyPath) {
  struct ppp_idle _ppp_idle;
  int fd = open(ttyPath, O_RDONLY);
  if (fd < 0) return 0;
  if (ioctl(fd, PPPIOCGIDLE, &_ppp_idle) < 0) goto err;
  close(fd);
  return optIdleBoot * 60 - min(_ppp_idle.xmit_idle, _ppp_idle.recv_idle);
err:  
  close(fd);
  return 0;
}

void PPPupdate(UserRec *ur, int unit) {
  struct ifreq _ifreq;
  struct ppp_stats _ppp_stats;
  int s;
  word sBytes;
  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) logQuit("couldn't create socket");
  bzero(&_ifreq, sizeof(struct ifreq));
  sprintf(_ifreq.ifr_ifrn.ifrn_name, "ppp%d", unit);
  _ifreq.ifr_ifru.ifru_data = (caddr_t)&_ppp_stats;
  if (ioctl(s, SIOCDEVPRIVATE, (caddr_t)&_ifreq) < 0) goto err;
  sBytes = _ppp_stats.p.ppp_ioctects + _ppp_stats.p.ppp_ooctects;
  if (sBytes > ur->PPPsBytes) {
    ur->PPPtBytes += sBytes - ur->PPPsBytes;
    ur->PPPsBytes = sBytes;
  }
err:
  close(s);
}

void PPPfindUnits() {
  int i, len, unit;
  DIR *d;
  FILE *f;
  pid_t pid;
  struct dirent *de;
  int unitStrLen;
  char unitStr[16], str[32];
  for (i = 0; i < MAX_PPP_UNITS; i++) PPPuser[i] = (uid_t)-1;
  if ((d = opendir("/var/run")) == NULL) logQuit("opendir failed");
  while ((de = readdir(d)) != NULL) {
    if ((len = strlen(de->d_name)) < 8) continue;
    if (strncmp(de->d_name, "ppp", 3)) continue;
    if (strcmp(de->d_name + strlen(de->d_name) - 3, "pid")) continue;
    for (i = 3, unitStrLen = 0; isdigit(de->d_name[i]); i++)
      unitStr[unitStrLen++] = de->d_name[i];
    unitStr[unitStrLen] = '\0';
    unit = atoi(unitStr);
    sprintf(str, "/var/run/ppp%d.pid", unit);
    if ((f = fopen(str, "r")) == NULL) continue;
    fgets(str, 32, f);
    pid = atoi(str);
    fclose(f);
    for (i = 0; i < nProcs; i++)
      if (!strcmp(procRec[i].name, "pppd") && procRec[i].pid == pid) {
	PPPuser[unit] = procRec[i].uid;
	break;
      }
  }
  closedir(d);
}

int PPPunit(uid_t uid) {
  int i;
  for (i = 0; i < MAX_PPP_UNITS; i++)
    if (PPPuser[i] == uid) return i;
  return -1;
}

void logQuit(char *format, ...) {
  char buf[256];
  va_list ap;
  va_start(ap, format);
  vsprintf(buf, format, ap);
  syslog(LOG_ERR, buf);
  exit(1);
}
